/* cache.c -- core routines for Coda MiniCache manipulation
   Copyright 1997-98 Michael Callahan (iget, iput)
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include <string.h>
#pragma intrinsic(strcmp, strcpy, strlen, memcpy, memcmp)

#include "MC.h"
#include "cache.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

#define CODA_HASH_SIZE 509

dllist_chain	cf_cache[CODA_HASH_SIZE]; /* index fid */


void init_cache(){
	int i=0;

	for ( i = 0 ; i <= CODA_HASH_SIZE-1 ; i++ )
		list_init(&(cf_cache[i]));
}

struct cnode *cnode_alloc(){
	struct cnode* cn = _HeapAllocate(sizeof(struct cnode), HEAPZEROINIT);
	if (!cn){
		LOGDEBUG("--cnode_alloc failed", 100);
		return NULL;
	}

	return cn;
}

void cnode_free(struct cnode *cn){
	if (!cn) return;

	print_cnode(cn);

	if (cn->c_container)
		_HeapFree(cn->c_container, 0);
	_HeapFree(cn, 0);
}

struct cnode *coda_lookupFid(struct ViceFid *fid){
	int hash = coda_hashfid(fid);	
	struct cnode *cn = NULL;
#ifdef CACHE
	struct dllist_head *lh, *le;
	print_fid(fid);
	le = &(cf_cache[hash]);
	lh = le;
	for ( le = lh->next ; le != lh ; le = le->next ) {
		cn = list_entry(le, struct cnode, c_cf);
		if ( coda_comparefid(&cn->c_fid, fid) ){
			LOGDEBUG("---c_lookupFid FOUND---", 10);
			goto out; 
		}
	}
	cn = NULL;
out:
#endif
	return cn;
}


int coda_revalidate(struct dentry *de){

	struct cnode *cn;

	LOGDEBUG("--coda_revalidate--", 1);

	if (!de) return 0;

	cn = (struct cnode *) de->d_vnode;

	if (!cn) return 0;

	printf("cnode flags %x", cn->c_flags);

	if (is_root(de)) return 1;

	if ( !(cn->c_flags & C_PURGE) /*&& VALID_VATTR(cn)*/)
		return 1;

	LOGDEBUG("coda_revalidate: invalid", 10);

	cn->c_flags &= ~(C_PURGE | C_VATTR | C_HAVEDATA | C_CONTAINER);

	return 0;
}


int coda_comparefid(struct ViceFid *f1, struct ViceFid *f2)
{   
	LOGDEBUG("--coda_comparefid", 1);
	if(!f1 || !f2) return 0;
	return ((f1->Unique == f2->Unique) && (f1->Vnode == f2->Vnode) &&
		(f1->Volume == f2->Volume));
	LOGDEBUG("--end of coda_comparefid", 1);
}

int coda_hashfid(struct ViceFid *fid){
	return ((fid->Unique ^ (fid->Vnode << 4)) % CODA_HASH_SIZE);
}


void coda_unflag_cnode(struct cnode *cn, int flag){
	if (!cn) return;

	cn->c_flags &= ~flag;
}

void coda_flag_cnode(struct cnode *cn, int flag){
	if (!cn) return;

	cn->c_flags |= flag;
}

void coda_flag_children(struct cnode *parent, int flag){
	struct dllist_head *child;
	struct dentry *de;

	child = parent->c_dentry->d_subdir.next;
	while ( child != &parent->c_dentry->d_subdir) {
		de = list_entry(child, struct dentry, d_child);
		child = child->next;
		/* don't know what to do with negative dentries */
		coda_flag_cnode((struct cnode *) de->d_vnode, flag);
	}
	return; 
}

void coda_replace_fid(struct cnode * cn, ViceFid *oldfid, ViceFid *newfid){
	if ( !coda_comparefid(&cn->c_fid, oldfid) ){
		LOGDEBUG("What? oldfid != cn->c_fid. Call 911", 10);
		return;
	}

	cn->c_fid = *newfid;

	list_del(&cn->c_cf);
	list_add(&cn->c_cf, &cf_cache[coda_hashfid(&cn->c_fid)]);
}

/* dcache ops */
 
int coda_delete(struct dentry *de){
	struct cnode *cn =  (struct cnode *)de->d_vnode;
	LOGDEBUG("--coda_delete", 1);
	if (cn->c_flags & C_PURGE)  /* cnode specific */
		d_drop(de);
	LOGDEBUG("--end of coda_delete", 1);
	return 0;
}

int coda_iput(struct dentry *de){
	struct cnode *cn = (struct cnode*)de->d_vnode;
	if (cn){ 
		list_del(&cn->c_cf); 
		list_init(&cn->c_cf);
	}
	return 0;
}

void coda_instantiate(struct dentry *de, struct cnode *cn, struct ViceFid *fid){
	LOGDEBUG("--coda_instantiate", 1);
	list_add(&cn->c_cf, &cf_cache[coda_hashfid(fid)]);
	cn->c_fid.Volume = fid->Volume;
	cn->c_fid.Vnode = fid->Vnode;
	cn->c_fid.Unique = fid->Unique;
	cn->c_dentry = de;
	d_instantiate(de, (DWORD) cn);
	LOGDEBUG("--end of coda_instantiate", 1);
}

int coda_ifree(struct dentry *de){
	struct cnode *cn = NULL;

	LOGDEBUG("--coda_ifree", 1);
	if (!de) return 1;

	cn = (struct cnode *) de->d_vnode;

	if (cn) cnode_free(cn);
	if (de->d_parent) {
		coda_unflag_cnode((struct cnode *)de->d_parent->d_vnode, C_HAVEDATA);
		((struct cnode*)de->d_parent->d_vnode)->c_search = 0;
	}
	
	LOGDEBUG("--end of coda_ifree", 1);
	return 0;
}

int coda_lookup(struct dentry **de, struct cstr *str, int flags){
	int res = 0;
	res = venus_lookup((*de)->d_parent, str, de, flags);
	return res;
}

