/* file.c -- Windows 95 IFS routines for manipulating open files
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include <string.h>
#pragma intrinsic (strcpy, strlen, memcpy, memcmp)

#include "mc.h"
#include <vmm.h>
#include "vmmextra.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static struct hndlmisc hf_tbl = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_HNDLMISC,

	ErrorStub2,
	ErrorStub2,
	ErrorStub2,
	ErrorStub2,			/* LockFile,*/
	ErrorStub2,			/* FileDateTime, */
	ErrorStub2,			/* NamedPipeRequest */
	ErrorStub2,			/* NamedPipeHandleInfo, */
	ErrorStub2			/* EnumerateHandle */
};


#define filesig 0xdeaffade

/* changed flag is important because: to set attr of a file, an 
application first has to open it with write permission. Closing
the file causes venus to store it, even when no writes have occured.
Opening a file with Powerpoint for example will cause multiple Stores 
to the servers. This is fatal with very large files. We only want Stores, 
when the file really was written. The changed flag is set in FS_WriteFile */
struct fileinfo {
	DWORD	signature;
	DWORD	memctx;
	DWORD	hdl;
	WORD	changed;
	DWORD	size;
	DWORD	flags;
	WORD		hidden;
	ParsedPath *ppath;
	struct cnode *cn;
};



int OpenFile (pioreq pir, struct dentry *de)
{
	DWORD takenaction, hdl = 0;
	WORD attr;
	int res = 0;
	int flags;
	struct cnode *cn = NULL;
	struct fileinfo *pfi = NULL;

	LOGDEBUG("--open--", 1);
	logucode (pir->ir_upath);
	logucode (pir->ir_uFName);

	pir->ir_fh = NULL;
	hdl = 0;

	if (!de) goto error;
	
	cn = (struct cnode*) de->d_vnode;

	if (cn && IS_INCONF(cn)){
		res = ERROR_ACCESS_DENIED;
		goto error;
	}

	if (!memcmp(de->d_name.name + de->d_name.len - 4, ".$cf", 4)){
		res = ERROR_ACCESS_DENIED;
		d_drop(de);
		goto error;
	}

	/* first, either look up inode or create new inode */

	/* first, if the request flags allow it, we try to open existing file */
	if (pir->ir_options & ACTION_EXISTS_OPEN) {
		LOGDEBUG("--open : ACTION_EXISTS_OPEN", 10);
		if (!res) goto gotit;
	}

	/* if we get here, then check if flags demand that we create a new file */
	if (pir->ir_options & ACTION_NEXISTS_CREATE){
		LOGDEBUG("--open : ACTION_NEXISTS_CREATE", 10);

		res = venus_create(pir, &de);
		cn = (struct cnode*) de->d_vnode;
	}
	
	if (res) goto error;	

gotit:


	flags = 0;
	switch (pir->ir_flags & ACCESS_MODE_MASK) {
	case ACCESS_READONLY:
	case ACCESS_EXECUTE:
		flags = C_O_READ;
		break;

	case ACCESS_WRITEONLY:
		flags = C_O_WRITE;
		break;

	case ACCESS_READWRITE:
		flags = C_O_READ | C_O_WRITE;
		break;

	default:
		printf ("Unknown access mode %x", pir->ir_flags & ACCESS_MODE_MASK);
	}
	if (pir->ir_options & ACTION_TRUNCATE)
		flags |= C_O_TRUNC;
	

	pfi = (struct fileinfo *) _HeapAllocate (sizeof(struct fileinfo), HEAPZEROINIT);
	if (!pfi) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		goto error;
	}


	printf("about to enter openVFid: Access: %x, Action: %x", (DWORD)(pir->ir_flags & ACCESS_MODE_MASK), (DWORD)(pir->ir_options & ACTION_MASK));
	res = openVFid(cn, flags, (WORD) (pir->ir_flags & ACCESS_MODE_MASK),
				(WORD) pir->ir_attr, (BYTE) (pir->ir_options & ACTION_MASK),
				/*req*/NULL, &hdl, &takenaction, &attr, /*NULL*/&pfi->hidden);
	if (res) {
		LOGDEBUG("open: OpenCreate failed", 10);
		goto error;
	}

	printf ("takenaction %x attr %x", takenaction, attr);

	pir->ir_options = (WORD) takenaction;
	pir->ir_attr = attr;

	// ir_size should get file size
	pir->ir_size = 0;
	if (pir->ir_options == ACTION_OPENED) {
		printf ("getfilesize (hdl %x)", hdl);
		res = R0_GetFileSize(hdl, &pir->ir_size);
		if (res) {
			LOGDEBUG("GetFileSize failed", 10);
			goto error;
		}
	}

	SetHandleFunc(pir, ReadFile, WriteFile, &hf_tbl);

	pfi->signature = filesig;
	pfi->memctx = (DWORD) _GetCurrentContext();
	pfi->hdl = hdl;
	pfi->changed = 0;
	pfi->flags = flags;
	pfi->cn = cn;
	pfi->size = pir->ir_size;
	pfi->ppath = pir->ir_ppath;

	pir->ir_fh = pfi;
	
error:
	if (res){
		if (hdl) R0_CloseFile(hdl);
		if (pfi) _HeapFree(pfi, 0);
		dput(de);
		printf ("open failed: %x", res);
	}
	else
		printf ("end open (%x) ctx %x", (int) pfi, _GetCurrentContext());

	LOGDEBUG("--end of open--", 1);

	return pir->ir_error = res;
}

void checkctx (struct fileinfo *pfi)
{
	if (pfi->memctx != (DWORD) _GetCurrentContext()) {
		printf ("CONTEXT ERROR: orig %x now %x",
			pfi->memctx, (DWORD) _GetCurrentContext());
	}
}

int ReadFile (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res;

	printf ("--read-- (%x) pos %x len %x addr %x", (int) pfi, pir->ir_pos, pir->ir_length, pir->ir_data);

	if (pfi->signature != filesig) {
		LOGDEBUG("BAD SIGNATURE!", 10);
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}

	checkctx (pfi);

	if (pir->ir_pos >= pfi->size)
		pir->ir_length = 0;
	else if (pir->ir_pos + pir->ir_length > pfi->size)
		pir->ir_length = pfi->size - pir->ir_pos;

	res = R0_ReadFile(FALSE,
			pfi->hdl,
			pir->ir_length,
			pir->ir_pos,
			(PBYTE) pir->ir_data,
			(PDWORD) &pir->ir_length);
	if (!res) {
		pir->ir_pos += pir->ir_length;
	}
	pir->ir_error = res;
	printf ("end read (err %x)", res);
	LOGDEBUG("--end of read--", 1);
	return pir->ir_error = res;
}

int WriteFile (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res;

	printf ("--write-- (%x) pos %x len %x", (DWORD) pfi, pir->ir_pos, pir->ir_length);

	if (pfi->signature != filesig) {
		LOGDEBUG("BAD SIGNATURE!", 10);
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}

	checkctx (pfi);

	res = R0_WriteFile(FALSE,
			pfi->hdl,
			pir->ir_length,
			pir->ir_pos,
			(PBYTE) pir->ir_data,
			(PDWORD) &pir->ir_length);
	if (!res) {
		pfi->changed = 1;
		pir->ir_pos += pir->ir_length;
		if (pir->ir_pos > pfi->size)
			pfi->size = pir->ir_pos;
	}

	pir->ir_error = res;
	printf ("end write (err %x)", res);
	LOGDEBUG("--end of write--", 1);
	return pir->ir_error = res;
}

int FileSeek (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = 0;

	printf ("--seek-- (%x) flags %x pos %x", 
		(int) pfi, pir->ir_flags, pir->ir_pos);

	if (pfi->signature != filesig) {
		LOGDEBUG("BAD SIGNATURE!", 10);
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx(pfi);

	if (pir->ir_flags == FILE_END) {
		DWORD dwSize;
		printf ("seek: about to do GetFileSize (hdl %x)", pfi->hdl);
		res = R0_GetFileSize (pfi->hdl,	&dwSize);
		if (!res) {
			pir->ir_pos += dwSize;
			printf ("seek: pos now %x", pir->ir_pos);
		}
	}
	
	printf ("--end of seek (err %x)--", res);
	return pir->ir_error = res;
}

int CloseFile (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = 0;

	printf ("--close-- (%x) closetype %x", (int) pfi, pir->ir_flags);
	if (pfi->signature != filesig) {
		LOGDEBUG("BAD SIGNATURE!", 10);
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx (pfi);

	if (pir->ir_flags != CLOSE_FINAL) {
		LOGDEBUG("not close final", 10);
		return pir->ir_error = res;
	}
	
	res = R0_CloseFile(pfi->hdl);
	/* XXX: what to do if there is an error here? */
	if (res !=0 ){
		printf ("Error at R0_CloseFile in CloseFile");
		LOGDEBUG(" -- close error ", 10);
	}

	printf("---CloseFile flags: %x ----", pfi->flags);

	if (!pfi->hidden){
		res = venus_close(pfi->cn->c_fid, pfi->flags);
		LOGDEBUG("FILE WAS CLOSED", 100);
		print_fid(&pfi->cn->c_fid);

	//	res = venus_close(pfi->cn->c_fid, (pfi->changed) ? pfi->flags : pfi->flags | C_O_CLEAN );
	}

	if (!res) /* adjust size in cache */
		pfi->cn->c_vattr.va_size = pfi->size;
error:
	dput (pfi->cn->c_dentry);
	_HeapFree (pfi, 0);

	printf ("--end of close (err %x)--", res);

	return pir->ir_error = res;
}

int CommitFile (pioreq pir)
{
	printf ("--commit-- (%x)", pir->ir_fh);
	LOGDEBUG("--end of commit--", 1);
	return pir->ir_error = 0;
}

int LockFile (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	printf ("--lock-- (%x) flags %x opt %x", (DWORD) pfi, pir->ir_flags, pir->ir_options);
	if (pfi->signature != filesig) {
		LOGDEBUG("BAD SIGNATURE!", 10);
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx (pfi);
	LOGDEBUG("--end of lock--", 1);
	return pir->ir_error = 0;
}

int FileDateTime (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = ERROR_INVALID_PARAMETER;

	printf ("--filedatetime-- (%x) subfunc %x", (DWORD) pfi, pir->ir_flags);
	if (pfi->signature != filesig) {
		LOGDEBUG("BAD SIGNATURE!", 10);
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx (pfi);

	res = coda_getattr(pfi->cn);
	if (res)
		goto error;
	
	switch (pir->ir_flags) {
	case SET_MODIFY_DATETIME:
		pfi->cn->c_vattr.va_mtime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		pfi->cn->c_vattr.va_mtime.ts_nsec = 0;
		break;

	case GET_MODIFY_DATETIME:
		pir->ir_dostime = IFSMgr_NetToDosTime(pfi->cn->c_vattr.va_mtime.ts_sec);
		break;

	case SET_LAST_ACCESS_DATETIME:
		pfi->cn->c_vattr.va_atime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		pfi->cn->c_vattr.va_atime.ts_nsec = 0;
		break;

	case GET_LAST_ACCESS_DATETIME:
		pir->ir_dostime = IFSMgr_NetToDosTime(pfi->cn->c_vattr.va_atime.ts_sec);
		break;

	case SET_CREATION_DATETIME:
		pfi->cn->c_vattr.va_ctime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		pfi->cn->c_vattr.va_ctime.ts_nsec = pir->ir_options * 10000000;
		break;

	case GET_CREATION_DATETIME:
		pir->ir_dostime = IFSMgr_NetToDosTime(pfi->cn->c_vattr.va_ctime.ts_sec);
		pir->ir_options = pfi->cn->c_vattr.va_ctime.ts_nsec / 10000000;
		break;
	}

	if (pir->ir_flags == SET_MODIFY_DATETIME ||
		pir->ir_flags == SET_LAST_ACCESS_DATETIME ||
		pir->ir_flags == SET_CREATION_DATETIME)
		res = coda_setattr(pfi->cn);

error:
	printf ("--end of filedatetime (err %x)--", res);

	return pir->ir_error = res;
}

int NamedPipeRequest (pioreq pir)
{
	printf ("FAILING -- NamedPipeRequest %x", pir->ir_flags);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}

int NamedPipeHandleInfo (pioreq pir)
{
	LOGDEBUG("FAILING -- NamedPipeHandleInfo", 1);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}

int EnumerateHandle (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = ERROR_INVALID_FUNCTION;

	LOGDEBUG("--enumeratehandle--", 1);

	checkctx (pfi);

	switch (pir->ir_flags) {
	case ENUMH_GETFILEINFO:
		LOGDEBUG("--enumeratehandle: getfileinfo", 10);
		res = coda_getattr(pfi->cn);
		if (res)
			break;			

		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_dwFileAttributes = (pfi->cn->c_vattr.va_type == C_VDIR) ? FILE_ATTRIBUTE_DIRECTORY : 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_ftCreationTime = IFSMgr_NetToWin32Time(pfi->cn->c_vattr.va_ctime.ts_sec);
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_ftLastAccessTime = IFSMgr_NetToWin32Time(pfi->cn->c_vattr.va_atime.ts_sec);
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_ftLastWriteTime = IFSMgr_NetToWin32Time(pfi->cn->c_vattr.va_mtime.ts_sec);
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_dwVolumeSerialNumber = 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileSizeHigh = 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileSizeLow = (unsigned long) pfi->cn->c_vattr.va_size;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nNumberOfLinks = (unsigned long) pfi->cn->c_vattr.va_nlink;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileIndexHigh = 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileIndexLow = 0;

		break;

	case ENUMH_GETFILENAME:
		LOGDEBUG("--enumeratehandle: getfilename", 10);
		if (pfi && pfi->ppath) 
			memcpy ((PBYTE) pir->ir_ppath, pfi->ppath, IFSPathSize(pfi->ppath));
		res = 0;
		break;

	case ENUMH_GETFINDINFO:
		LOGDEBUG("--enumeratehandle: getfindinfo", 10);
		if (pfi && pfi->ppath) 
			memcpy ((PBYTE) pir->ir_ppath, pfi->ppath, IFSPathSize(pfi->ppath));
		res = 0;
		break;

	case ENUMH_RESUMEFIND:
		LOGDEBUG("--enumeratehandle: resumefind", 10);
		break;

	case ENUMH_RESYNCFILEDIR:
		LOGDEBUG("     resyncfiledir", 10);
		break;

	default:
		LOGDEBUG("     unknown subfunc code!", 10);
		break;
	}

	LOGDEBUG("--end of enumeratehandle--", 1);

	return pir->ir_error = res;
}

