/* ipc.c -- IPC interface with Venus
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/


#include "mc.h"
#include "vwin32wrap.h"

#include "vmmextra.h"

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG

/* list of requests waiting to be picked up by a potemkin
	thread */
static struct req_list_entry pending = {&pending, &pending};

/* list of requests currently being processed by a potemkin
	thread */
static struct req_list_entry processing = {&processing, &processing};

/* pmtx: protects modifications to above two lists */
static PVMMMUTEX pmtx = NULL;		

/* post: pointer to DWORD to set to 1 and to signal (using SignalID) 
   when a request is available -- venus/potemkin will use WaitOnID to pause
   (e.g. in SOCK.VXD's sys_select()) */
PDWORD post = NULL;

/* queue_empty -- return whether queue is empty */
BOOL queue_empty (struct req_list_entry *list)
{
	return list->next == list;
}

/* enqueue/dequeue -- must be called while holding the pmtx
   list-protection mutex */

void enqueue (struct req_list_entry *list,
			  struct request *req)
{
	struct req_list_entry *le = &(req->le);
	list->prev->next = le;
	le->prev = list->prev;
	le->next = list;
	list->prev = le;
}

struct request *dequeue (struct req_list_entry *list)
{
	struct req_list_entry *le;

	if (list->next == list)
		return NULL;  /* empty list */

	le = list->next;
	list->next = le->next;
	le->next->prev = list;

	return (struct request *) le;
}


static void remove (struct request *req)
{
	struct req_list_entry *le = (struct req_list_entry *) req;

	le->prev->next = le->next;
	le->next->prev = le->prev;
}


struct request *dequeue_by_id (struct req_list_entry *list,
							   unsigned long unique)
{
	struct req_list_entry *le;

	for (le = list->next; le != list; le = le->next) {
		if (((struct request *) le)->r.in.ih.unique == unique)
			break;
	}

	if (le == list)
		return NULL;   /* not there */

	le->prev->next = le->next;
	le->next->prev = le->prev;
	return (struct request *) le;
}



/* initialize_ipc -- initializes IPC system */

int initialize_ipc ()
{
	static int needs_init = 1;

	if (!needs_init)
		return 0;

	pmtx = _CreateMutex(0, 0);
	if (!pmtx) {
		LOGDEBUG("error making pmtx", 10);
		return ERROR_NO_MORE_FILES;
	}

	needs_init=0;
	return 0;
}

/* process_rpc -- called by thread wishing to invoke an RPC
   to a potemkin.  blocks until reply arrives */

static unsigned long unique = 1;

int process_rpc (struct request *req)
{
	if (!pmtx)
		return ERROR_INVALID_FUNCTION;

	/* set flag for waiting */
	req->doneflag = 0;

	/* add request to pending list */
	_EnterMutex (pmtx,  BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);
	req->r.in.ih.unique = unique++;
	enqueue (&pending, req);
	_LeaveMutex (pmtx);

	/* signal waiting potemkin that there is a request */
	if (post) {
		*post = 1;
		_SignalID ((DWORD) post);
	}

	/* wait for our request to be satisfied */
	printf ("CLIENT: %x: enqueue (func %x) ", req->r.in.ih.unique, req->r.in.ih.opcode);
	while (!req->doneflag) {
		_asm pushf
		_asm cli
		if (!req->doneflag)
			_BlockOnID ((DWORD) &req->doneflag, BLOCK_THREAD_IDLE | BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);   /* XXX */
		_asm popf
	}
	printf ("CLIENT: %x: done (err %x)", req->r.out.oh.unique, req->r.out.oh.result);

	/* when we return here, we have been signalled */
	return 0;
}


/* insert_upcall -- insert an upcall in the pending queue */
int insert_upcall (struct request *req)
{
	if (!pmtx)
		return ERROR_INVALID_FUNCTION;

	/* add request to pending list */
	_EnterMutex (pmtx,  BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);
	req->doneflag = 0;
	req->r.in.ih.unique = unique++;
	enqueue (&pending, req);
	_LeaveMutex (pmtx);

	/* signal any waiting potemkins that there is a request */
	printf ("UPCALL: %x", req->r.in.ih.unique);
	if (post) {
		*post = 1;
		_SignalID ((DWORD) post);
	}

	return 0;
}

#define POLL_READ  0x00000001

DWORD do_poll (PVOID context, PDWORD setpost)
{
	post = setpost;
	if (!queue_empty(&pending))
		return POLL_READ;
	return 0;
}

void do_close (PVOID context)
{
}

/* get_pending -- called by potemkin thread to get request */
struct request *get_pending (int blockflag)
{
	DWORD postval, res;
	struct request *req;

	do {
		if (blockflag) {
			postval = 0;
			res = do_poll (NULL, &postval);
			if (!(res & POLL_READ)) {
				printf ("SERVER: blocking on %x", (DWORD) &postval);
				_asm pushf
				_asm cli
				if (!postval)
					_BlockOnID ((DWORD) &postval, BLOCK_THREAD_IDLE | BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);
				_asm popf
			}
			do_poll (NULL, NULL);
		}

		_EnterMutex (pmtx,  BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);
		req = dequeue (&pending);

		/* add to processing list, unless it is an upcall (func => 100) */
		if (req && req->r.in.ih.opcode < 100)
			enqueue (&processing, req);

		_LeaveMutex (pmtx);
	} while (blockflag && !req);

	if (req)
		printf ("SERVER: %x: got request (func %x)", req->r.in.ih.unique, req->r.in.ih.opcode);
	else
		LOGDEBUG("SERVER: no request!", 1);
	return req;
}


/* find_request -- called by potemkin to complete request */
struct request *find_request (long unique)
{
	struct request *req;
	_EnterMutex (pmtx,  BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);
	req = dequeue_by_id (&processing, unique);
	_LeaveMutex (pmtx);
	if (req)
		printf ("SERVER: %x: ready to answer", req->r.in.ih.unique);
	else
		printf ("SERVER: %x: ERROR--no such request", unique);
	return req;
}

/* signal_request -- called by potemkin after copying reply
   back in */
void signal_request (struct request *req)
{
	req->doneflag = 1;
	_SignalID ((DWORD) &req->doneflag);
}

static void clear_queue_by_rh (struct req_list_entry *list, DWORD rh)
{
	struct req_list_entry *le, *le2;
	struct request *req;

	le = list->next;
	while (le != list) {
		if (((struct request *) le)->rh == rh) {
			/* first, jump off this element */
			le2 = le;
			le = le->next;

			req = (struct request *) le2;
			remove(req);

			req->r.out.oh.result = EIO;
			signal_request (req);
		} else {
			le = le->next;
		}
	}
}

void clear_resource_requests (DWORD rh)
{
	_EnterMutex (pmtx, BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);

	clear_queue_by_rh (&pending, rh);
	clear_queue_by_rh (&processing, rh);

	_LeaveMutex (pmtx);
}

static void age_queue (struct req_list_entry *list, DWORD agelimit)
{
	struct req_list_entry *le, *le2;
	struct request *req;

	le = list->next;
	while (le != list) {
		if (++((struct request *) le)->age >= (int) agelimit) {
			/* first, jump off this element */
			le2 = le;
			le = le->next;

			req = (struct request *) le2;
			remove(req);
			req->r.out.oh.result = EIO;
			signal_request (req);
		} else {
			le = le->next;
		}
	}
}

void heartbeat()
{
	PTCB lockthr;
    int gotit = 0;

    _asm pushf
    _asm cli
    lockthr = _GetMutexOwner (pmtx);
    if (lockthr) {
		_asm popf;
		return;
	}
    _EnterMutex (pmtx, 0);
    _asm popf

	age_queue (&pending, 2);
	age_queue (&processing, 10);
//	age_queue (&pending, 2);
//	age_queue (&processing, 2);

	_LeaveMutex (pmtx);
}

